/*
 * Copyright Dept. of Mathematics & Computer Science Univ. Paris-Descartes
 *
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

package pddl4j.forward;

import java.io.File;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import pddl4j.ErrorManager;
import pddl4j.InvalidExpException;
import pddl4j.PDDLObject;
import pddl4j.Parser;
import pddl4j.RequireKey;
import pddl4j.Source;
import pddl4j.ErrorManager.Message;
import pddl4j.exp.AndExp;
import pddl4j.exp.AtomicFormula;
import pddl4j.exp.Exp;
import pddl4j.exp.ExpID;
import pddl4j.exp.InitEl;

/**
 * This class implements the forward planner as an example of the pddl4j
 * library. Only strips and typing are accepted.
 *
 * @author Damien Pellier
 * @version 1.0
 */
public final class Forward {

    /**
     * The problem to solve.
     */
    private PDDLObject problem;

    /**
     * The goals.
     */
    private BitSet goals;

    /**
     * The initial state.
     */
    private BitSet init;

    /**
     * The list of operators of the planning problem.
     */
    private List<Op> ops_table;

    /**
     * The list of facts of the planning problem.
     */
    private List<Fact> facts_table;

    /**
     * The size of the facts table.
     */
    private int facts_table_size;

    /**
     * The size of the operators table.
     */
    private int ops_table_size;

    /**
     * The list of solution plans found.
     */
    private List<Plan> plans;

    /**
     * The number of solution plan to be found.
     */
    private int nbPlans;

    /**
     * Creates a new java forward planner to solve a specific problem.
     *
     * @param problem the problem to solve.
     */
    private Forward(final PDDLObject problem) {
        this.ops_table = new ArrayList<Op>(100);
        this.facts_table = new ArrayList<Fact>(100);
        this.problem = problem;
        this.plans = new LinkedList<Plan>();
        this.nbPlans = 1;
    }


    /**
     * The main method the graphplan planner.
     *
     * @param args an array of string representing the path of the planning
     *            domain and the problem.
     */
    public static void main(final String[] args) {
        try {
            // Checks the command line
            if (args.length != 2) {
                Forward.printUsage();
            } else {
                // Gets the pddl compiler options
                final Properties options = Forward.getParserOptions();
                // Creates an instance of the java pddl compiler
                if (!new File(args[0]).exists()) {
                    System.out.println("domain file " + args[0] + " does not exist");
                    System.exit(0);
                }
                if (!new File(args[1]).exists()) {
                    System.out.println("problem file " + args[1] + " does not exist");
                    System.exit(0);
                }
                final Parser parser = new Parser(options);
                final PDDLObject domain = parser.parse(new File(args[0]));
                final PDDLObject problem = parser.parse(new File(args[1]));
                PDDLObject pb = null;
                if (domain != null && problem != null) {
                    pb = parser.link(domain, problem);
                }
                // Gets the error manager of the pddl compiler
                final ErrorManager mgr = parser.getErrorManager();
                // If the compilation produces errors we print it and stop
                if (mgr.contains(Message.ERROR)) {
                    mgr.print(Message.ALL);
                }
                // else we print the warning and start the planning process
                else {

                    mgr.print(Message.WARNING);

                    System.out.println("\nparsing domain \"" + domain.getDomainName() + "\" done successfully ...");
                    System.out.println("parsing problem \"" + problem.getProblemName() + "\" done successfully ...\n");

                    final Forward planner = new Forward(pb);
                    planner.preprocessing();
                    System.out.println("preprocessing ...\n");
                    System.out.printf("number of instantiated actions : %12d \n", planner.ops_table_size - planner.facts_table_size);
                    System.out.printf("number of instantiated facts   : %12d \n\n", planner.facts_table_size);

                    System.out.println(planner.solve());
                }
            }

        } catch (Throwable t) {
            System.err.println(t.getMessage());
            t.printStackTrace(System.err);
        }
    }



    /**
     * Returns the default options of the compiler.
     *
     * @return the default options.
     */
    private static Properties getParserOptions() {
        final Properties options = new Properties();
        options.put("source", Source.V3_0);
        options.put(RequireKey.STRIPS, true);
        options.put(RequireKey.TYPING, true);
        options.put(RequireKey.EQUALITY, true);
        options.put(RequireKey.NEGATIVE_PRECONDITIONS, false);
        options.put(RequireKey.DISJUNCTIVE_PRECONDITIONS, false);
        options.put(RequireKey.EXISTENTIAL_PRECONDITIONS, false);
        options.put(RequireKey.UNIVERSAL_PRECONDITIONS, false);
        options.put(RequireKey.CONDITIONAL_EFFECTS, false);
        return options;
    }


    /**
     * Print the usage of the forward command.
     */
    private static void printUsage() {
        System.out.println("Usage: forward [domain] [problem]");
        System.exit(0);
    }

    //-----------------------------------------------------------------------//
    // Methods used for search a plan
    // You have to complete this method
    //-----------------------------------------------------------------------//

    /**
     * Solves the planning problem and returns the first solution plan found.
     * This method must be completed.
     *
     * @return the first solution plan found or failure if
     *         no solution was found.
     */
    public boolean solve() {

    }

    /**
     * Returns <code>true</code> if an operator is applicable in a specified
     * state. An operator is applicable to state its preconditions are included
     * in the state.
     *
     * @param op The operator to be tested.
     * @param state the state bit set.
     * @return <code>true</code> if an operator is applicable to a specified
     *         state; <code>false</code> otherwise.
     */
    private boolean isApplicable(final Op op, BitSet state) {
        final BitSet pre = op.getPositivePreconditon();
        return this.contains(pre, state);
    }

    /**
     * Returns <code>true</code> if a specified set of fact is contains in a
     * specific level of the graph.
     *
     * @param facts the facts to be tested.
     * @param state the level of the graph where the test is done.
     * @return <code>true</code> if a specified set of fact is contains in a
     *         specific level of the graph; <code>false</code>otherwise.
     */
    public boolean contains(final BitSet facts, final BitSet state) {
        final BitSet cs = (BitSet) state.clone();
        cs.and(facts);
        return cs.equals(facts);
    }

    //-----------------------------------------------------------------------//
    // Methods used for planning pre-processing                               //
    //-----------------------------------------------------------------------//

    /**
     * Executes the pre-processing of the problem, i.e, instantiates the
     * operators of the problem, computes the static dependence between
     * operators, initialize the first level of the planning graph and the goal.
     *
     * @throws InvalidExpException if an PDDL expression of the goal or in an
     *             action definition is not accepted by the planner
     */
    private void preprocessing() throws InvalidExpException {
        this.facts_table = Utilities.instantiateFacts(this.problem);
        this.facts_table_size = this.facts_table.size();
        this.ops_table = Utilities.instantiateOps(this.problem, this.facts_table);
        this.ops_table_size = this.ops_table.size();
        this.goals = this.initGoals(this.problem.getGoal(), this.facts_table);
        this.init = this.initInitialState(this.problem.getInit(), this.facts_table);
    }

    /**
     * Initialize the planning graph, i.e., create the first fact level of the
     * planning graph.
     *
     * @param init The initial state under the form of a list of predicates.
     * @param facts the instantiated facts.
     * @return The initial state bit set.
     * @throws InvalidExpException if an invalid PDDL expression is found in the
     *             PDDL initial state description of the problem.
     */
    public BitSet initInitialState(final List<InitEl> init,
                final List<Fact> facts) throws InvalidExpException {
        final BitSet is = new BitSet(facts.size());
        for (InitEl el : init) {
            if (el.getExpID().equals(ExpID.ATOMIC_FORMULA)) {
                final AtomicFormula predicate = (AtomicFormula) el;
                is.set(facts.indexOf(predicate), true);
            } else {
                throw new InvalidExpException(el.getExpID());
            }
        }
        return is;
    }

    /**
     * Initialize the goals of the problem, i.e., converts the goal of the in a PDDL
     * format to inner format.
     *
     * @param goal The goal under the form of a list of predicates.
     * @param facts the instantiated facts.
     * @return The goal bit set.
     *
     * @throws InvalidExpException if an invalid PDDL expression is found in
     *             the PDDL goal description of the problem.
     */
    public BitSet initGoals(final Exp goal, final List<Fact> facts) throws InvalidExpException {
        BitSet g = new BitSet(facts.size());
        if (goal.getExpID().equals(ExpID.AND)) {
            final AndExp andExp = (AndExp) goal;
            for (Exp e : andExp) {
                if (e.getExpID().equals(ExpID.ATOMIC_FORMULA)) {
                    final AtomicFormula predicate = (AtomicFormula) e;
                    g.set(facts.indexOf(predicate), true);
                } else {
                    throw new InvalidExpException(e.getExpID());
                }
            }
        } else {
            throw new InvalidExpException(goal.getExpID());
        }
        return g;
    }

    //-----------------------------------------------------------------------//
    // Methods used to print                                                 //
    //-----------------------------------------------------------------------//

    /**
     * Print the ops table.
     */
    private void printOpsTable() {
        for (Op op : this.ops_table) {
            System.out.println(this.opToString(op));
        }
    }

    /**
     * Returns a string representation of a specified operator.
     *
     * @param op the operator to convert.
     * @return the string representation of the specified operator.
     * @see java.lang.Object#toString()
     */
    public String opToString(final Op op) {
        final StringBuffer str = new StringBuffer();
        str.append("Op ");
        str.append(op.getName());
        str.append("\n");
        str.append("precondition:\n");
        str.append("(");
        if (!op.getPositivePreconditon().isEmpty()) {
            int i =op.getPositivePreconditon().nextSetBit(0);
            if (i >= 0) {
                str.append(this.facts_table.get(i));
            }
            i = op.getPositivePreconditon().nextSetBit(i + 1);
            while (i >= 0) {
                str.append(" AND ");
                str.append(this.facts_table.get(i));
                i = op.getPositivePreconditon().nextSetBit(i + 1);
            }
        }
        if (!op.getNegativePrecondition().isEmpty()) {
            int i = op.getNegativePrecondition().nextSetBit(0);
            if (i >= 0) {
                str.append(this.facts_table.get(i));
            }
            i = op.getNegativePrecondition().nextSetBit(i + 1);
            while (i >= 0) {
                str.append(" AND (NOT ");
                str.append(this.facts_table.get(i));
                str.append(")");
                i = op.getNegativePrecondition().nextSetBit(i + 1);
            }
        }
        str.append(")\n");
        str.append("effect:\n");
        str.append("(");
        if (!op.getPositiveEffect().isEmpty()) {
            int i = op.getPositiveEffect().nextSetBit(0);
            if (i >= 0) {
                str.append(this.facts_table.get(i));
            }
            i = op.getPositiveEffect().nextSetBit(i + 1);
            while (i >= 0) {
                str.append(" AND ");
                str.append(this.facts_table.get(i));
                i = op.getPositiveEffect().nextSetBit(i + 1);
            }
        }
        if (!op.getNegativeEffect().isEmpty()) {
            int i = op.getNegativeEffect().nextSetBit(0);
            if (i >= 0) {
                str.append(this.facts_table.get(i));
            }
            i = op.getNegativeEffect().nextSetBit(i + 1);
            while (i >= 0) {
                str.append(" AND (NOT ");
                str.append(this.facts_table.get(i));
                str.append(")");
                i = op.getNegativeEffect().nextSetBit(i + 1);
            }
        }
        str.append(")");
        return str.toString();

    }
}
